from pyglet import gl


def clamp(value, minimum=0, maximum=1):
    if value < minimum:
        return minimum
    if value > maximum:
        return maximum
    return value


class reify:
    def __init__(self, func):
        self.func = func

    def __get__(self, obj, cls):
        if obj is None:
            return self
        value = self.func(obj)
        setattr(obj, self.func.__name__, value)
        return value


def draw_rect(w, h):
    gl.glBegin(gl.GL_TRIANGLE_FAN)
    gl.glVertex2f(0, 0)
    gl.glVertex2f(0, h)
    gl.glVertex2f(w, h)
    gl.glVertex2f(w, 0)
    gl.glVertex2f(0, 0)
    gl.glEnd()


def draw_rect_xy(x, y, w, h, mode=gl.GL_TRIANGLE_FAN, last=True):
    gl.glBegin(mode)
    gl.glVertex2f(x+w, y)
    gl.glVertex2f(x, y)
    gl.glVertex2f(x, y+h)
    gl.glVertex2f(x+w, y+h)
    if last:
        gl.glVertex2f(x+w, y)
    gl.glEnd()


def draw_keyhole(xywh):
    x, y, w, h = xywh
    gl.glBegin(gl.GL_TRIANGLE_FAN)
    gl.glVertex2f(x, y-w)
    gl.glVertex2f(x, y+h+w)
    gl.glVertex2f(x+w, y+h)
    gl.glVertex2f(x+w, y)
    gl.glVertex2f(x, y-w)
    gl.glEnd()


def interp(a, b, t):
    t = clamp(t, 0, 1)
    return a * (1-t) + b * t



import types
import random

import pyglet

from jrti_game.data import default_keymap


state = types.SimpleNamespace(
    zoom=1,
    target_zoom=1,
    alt_zoom_start=None,
    last_drag_pos=None,
    center=[0, 0],
    last_mouse_pos=(400, 300),
    time=0,
    tool=None,
    tool_x=0,
    tool_y=0,
    reassign_state=None,
    reassign_anim_start=0,
    keymap={getattr(pyglet.window.key, k): v
            for k, v in default_keymap.items()},
    code=[random.randrange(10) for i in range(4)],
    key_config=[0, 3, 4, 0],
    stuck_keys=[],
    ending=False,
    min_zoom=1,
    end_time=None,
    zoom_easing=1,
)



import random
import contextlib
import math

from pyglet import gl
import pyglet.text
from attr import attrs, attrib

from jrti_game.data import spritesheet_texture, spritesheet_data
from jrti_game.data import instruction_font_name
from jrti_game.instructions import get_instruction_text
from jrti_game.util import clamp, reify, draw_rect, draw_keyhole
from jrti_game.state import state

tau = math.pi*2


@attrs(repr=False)
class Flippable:
    transparent = False

    x = attrib(0)
    y = attrib(0)
    width = attrib(8)
    height = attrib(8)
    scale = attrib(1)
    parent = attrib(None)
    color = attrib((1, 1, 1))
    bgcolor = attrib((0, 0, 0, 1))
    instructions_color=attrib((0.4, 0.9, 1))
    instructions = attrib('None')
    margin = attrib(1)

    drag_info = None, None
    flip_params = None
    flip_direction = None
    open_child_counter = 0
    close_speed = 0
    light_seed = 0

    hides_bugs = False
    num_hiding_bugs = 0
    last_moved = 0

    keyhole = None
    unlocked = None
    keyhole_params = 10, 400
    flip_locked = False

    children = attrib(convert=list)

    @children.default
    def _default(self):
        return []

    def __attrs_post_init__(self):
        if self.parent:
            self.parent.add_child(self)

    def __repr__(self):
        return '<{}: {}>'.format(type(self).__name__, self.instructions)

    def draw(self, zoom, children=True, **kwargs):
        if children:
            for child in self.children:
                child.draw_outer(zoom=zoom*child.scale, **kwargs)

    def tick(self, dt):
        if (self.flip_params and
                self.drag_info[0] is not self and
                not self.flip_locked):
            self.close_speed += (dt * 500000 / (self.width * self.height)**.5 /
                                 self.scale / state.zoom)
            angle = dt * self.close_speed
            x, y, rx, ry = self.flip_params
            if abs(rx) < angle and abs(ry) < angle:
                obj = self
                if obj is self:
                    while obj:
                        obj.open_child_counter -= 1
                        obj = obj.parent
                self.flip_params = None
            else:
                if rx > 0:
                    rx -= angle
                elif rx < 0:
                    rx += angle
                elif ry > 0:
                    ry -= angle
                elif ry < 0:
                    ry += angle
                self.flip_params = x, y, rx, ry

        for child in self.children:
            child.tick(dt)

    def add_child(self, child):
        self.children.append(child)
        child.parent = self

    def die(self):
        self.parent.children.remove(self)
        self.parent = None

    def spawn_bug(self, *args, **kwargs):
        pass

    def grab_move(self, dx, dy):
        pass

    def mouse_press(self, x, y, **kwargs):
        if self.drag_info[0] or self.unlocked:
            self.mouse_release()
        self.drag_start(x, y, **kwargs)
        self.mouse_drag(x, y, **kwargs)

    def mouse_drag(self, x, y, *, zoom=1, **kwargs):
        obj, start = self.drag_info
        if obj is self:
            pass
        elif obj is not None:
            obj.mouse_drag((x - obj.x) / obj.scale, (y - obj.y) / obj.scale,
                           zoom=zoom * obj.scale,
                           **kwargs)
            return
        else:
            return

        if self.drag_info[1]:
            sx, sy = self.drag_info[1]
            dx = abs(sx - x)
            dy = abs(sy - y)
            if self.flip_direction is None:
                if dx > dy:
                    direction = sx - x, 0
                else:
                    direction = 0, sy - y
                if dx > 4/zoom or dy > 4/zoom:
                    self.flip_direction = direction
            else:
                direction = self.flip_direction
            if direction[0]:
                mouse = x
                start = sx
                dim = self.width
                vdir = direction[0]
            else:
                mouse = y
                start = sy
                dim = self.height
                vdir = direction[1]
            if vdir < 0 and mouse < start:
                mouse = start
            if vdir > 0 and mouse > start:
                mouse = start
            if mouse < start:
                axis = 0
                factor = -1
            else:
                axis = dim
                factor = 1
            if start == axis:
                angle = 0
            elif mouse == axis:
                angle = 90 * factor
            else:
                ratio = ((mouse - axis) / (start - axis))
                ratio = clamp(ratio, -1, 1)
                angle = math.degrees(math.acos(ratio) * factor)

            z = zoom
            if abs(angle) > 5 and (
                    z*self.width >= 800 or z*self.height >= 600
                    or self.unlocked):
                self.mouse_release()
                if self.unlocked:
                    self.unlocked.anim_start = state.time
                angle = clamp(angle, -5, 5)

            if direction[0]:
                self.flip_params = axis, 0, -angle, 0
            else:
                self.flip_params = 0, axis, 0, -angle

    def mouse_release(self, **kwargs):
        obj, start = self.drag_info
        if obj and obj is not self:
            obj.mouse_release()
        self.flip_direction = None
        self.drag_info = None, None
        self.close_speed = 0

    def hit_test(self, x, y):
        return (0 <= x < self.width) and (0 <= y < self.height)

    def hit_test_grab(self, x, y):
        return False

    def hit_test_slab(self, x, y, w, h):
        for ex in range(int(x), int(x) + int(math.ceil(w)) or 1):
            for ey in range(int(y), int(y) + int(math.ceil(h)) or 1):
                if self.hit_test(ex, ey) and self.hit_test_grab(ex, ey):
                    return True
        return False

    def hit_test_overlap(self, x, a, b, mode):
        return -100

    def hit_test_all(self, x, y):
        if self.hit_test(x, y):
            for child in reversed(self.children):
                yield from child.hit_test_all((x - child.x) / child.scale,
                                              (y - child.y) / child.scale)
            yield self, x, y

    def draw_grab(self, **kwargs):
        parents = []
        obj = self
        while obj:
            parents.append(obj)
            obj = obj.parent
        with contextlib.ExitStack() as stack:
            for obj in reversed(parents):
                stack.enter_context(obj.draw_context(bg=False))
            if self.bgcolor[:3] == (0, 0, 0):
                gl.glColor4f(0.0, 0.2, 0.1, 1 / 3)
                draw_rect(self.width, self.height)
            self.draw_outline((0.1, 0.9, 0.4), 0.7)

    def drag_start(self, x, y, zoom, **kwargs):
        self.flip_locked = False
        for child in reversed(self.children):
            cx = (x - child.x) / child.scale
            cy = (y - child.y) / child.scale
            if child.hit_test(cx, cy):
                child.mouse_press(cx, cy, zoom=zoom*child.scale, **kwargs)
                self.drag_info = child, (x, y)
                return

        self.drag_info = self, (x, y)
        self.flip_direction = None
        self.light_seed = random.randint(0, 2**32)

        if state.tool.grabbing is self:
            state.tool.deactivate()

        obj = self
        while obj:
            obj.open_child_counter += 1
            obj = obj.parent

    @contextlib.contextmanager
    def draw_context(self, bg=True, flipping=False, **kwargs):
        gl.glPushMatrix()
        try:
            gl.glTranslatef(self.x, self.y, 0)
            gl.glScalef(self.scale, self.scale, self.scale)

            if flipping and self.flip_params:
                x, y, rx, ry = self.flip_params
                gl.glTranslatef(x, y, 0)
                gl.glRotatef(-ry, 1, 0, 0)
                gl.glRotatef(rx, 0, 1, 0)
                gl.glTranslatef(-x, -y, 0)

            if bg and self.bgcolor:
                gl.glColor4f(*self.bgcolor)
                draw_rect(self.width, self.height)

            gl.glColor3f(*self.color)

            yield
        finally:
            gl.glPopMatrix()

    def draw_outer(self, **kwargs):
        with self.draw_context(bg=False, **kwargs):
            if not self.flip_params:
                self.draw(**kwargs)
                if self.unlocked:
                    self.unlocked.draw_stuck()

    def draw_flipping(self, zoom, **kwargs):
        if self.flip_params:
            with self.draw_context(zoom=zoom, **kwargs):
                self.draw_light(zoom=zoom, **kwargs)
                self.draw_instructions(zoom=zoom, **kwargs)
            with self.draw_context(flipping=True, bg=True, zoom=zoom,
                                   **kwargs):
                self.draw(zoom=zoom, **kwargs)
                if self.unlocked:
                    self.unlocked.draw_stuck()
                self.draw_outline(zoom=zoom, **kwargs)

                gl.glEnable(gl.GL_CULL_FACE)
                gl.glColor4f(*self.instructions_color,
                             clamp(random.normalvariate(3/4, 1/120), 0, 1))
                draw_rect(self.width, self.height)
                gl.glDisable(gl.GL_CULL_FACE)

        if self.children and self.open_child_counter:
            with self.draw_context(bg=False, flipping=self.flip_params,
                                   zoom=zoom):
                for obj in self.children:
                    obj.draw_flipping(zoom=zoom*obj.scale, **kwargs)

    def draw_instructions(self, **kwargs):
        gl.glColor4f(*self.instructions_color,
                     clamp(random.normalvariate(1, 1/120), 0, 1))
        draw_rect(self.width, self.height)

        gl.glColor3f(0, 0, 0)
        if self.instruction_label:
            scale = self.instruction_label._jrti_scale
            gl.glScalef(1/scale, 1/scale, 1)
            self.instruction_label.draw()
        gl.glBindTexture(gl.GL_TEXTURE_2D, spritesheet_texture.id)

    @reify
    def instruction_label(self):
        scale = 1
        while self.width * scale < 400 and self.height * scale < 300:
            scale *= 1.1
        document = pyglet.text.document.UnformattedDocument(
            get_instruction_text(self.instructions)
        )
        document.set_style(0, 0, {
            'color': (0, 0, 0, 255),
            'font_name': instruction_font_name,
            'font_size': 25,
        })
        if self.width < self.margin*3 or self.height < self.margin*3:
            return None
        layout = pyglet.text.layout.TextLayout(
            document,
            width=(self.width - self.margin*2) * scale,
            height=(self.height - self.margin*2) * scale,
            multiline=True,
        )
        layout.x = self.margin * scale
        layout.y = self.margin * scale
        for size in range(25, 1, -1):
            if (layout.content_width > layout.width or
                    layout.content_height > layout.height):
                document.set_style(0, 0, {'font_size': size})
        for size in range(10, 1, -1):
            if (layout.content_width > layout.width or
                    layout.content_height > layout.height):
                document.set_style(0, 0, {'font_size': size/10})
        layout._jrti_scale = scale
        return layout

    def update_instructions(self):
        try:
            del self.instruction_label
        except AttributeError:
            pass
        for child in self.children:
            child.update_instructions()

    def draw_light(self, **kwargs):
        if self.flip_params:
            x, y, rx, ry = self.flip_params
            angle = abs(rx + ry)
            if angle < 90:
                wph = self.width + self.height
                N = 80 / (angle+1) * 20 * self.scale + 2
                if N > 50:
                    N = 50
                r = random.Random()
                r.seed(self.light_seed)
                for i in range(int(N)):
                    p = (angle / 60 / N * i) ** 1.1 / 5
                    diag_angle = tau/8
                    diag_angle += r.normalvariate(0, tau/16)
                    diag_angle += random.normalvariate(0, tau/256)
                    pw = p * wph * math.cos(diag_angle)
                    ph = p * wph * math.sin(diag_angle)
                    points = self.get_light_points(rx, ry, pw, ph)
                    gl.glColor4f(*self.instructions_color,
                                 (1 - angle / 60) / N * 2)
                    gl.glBegin(gl.GL_TRIANGLE_FAN)
                    for point in points:
                        gl.glVertex2f(*point)
                    gl.glVertex2f(*points[0])
                    gl.glEnd()

    def get_light_points(self, rx, ry, pw, ph):
        w = self.width
        h = self.height
        if rx:
            if rx > 0:
                return [
                    (0, h),
                    (w+pw, h+ph),
                    (w+pw, -ph),
                    (0, 0),
                ]
            else:
                return [
                    (w, 0),
                    (-pw, -ph),
                    (-pw, h+ph),
                    (w, h),
                ]
        elif ry:
            if ry > 0:
                return [
                    (0, 0),
                    (-pw, h+ph),
                    (w+pw, h+ph),
                    (w, 0),
                ]
            else:
                return [
                    (w, h),
                    (w+pw, -ph),
                    (-pw, -ph),
                    (0, h),
                ]
        else:
            return [
                (-pw, -ph),
                (-pw, h+ph),
                (w+pw, h+ph),
                (w+pw, -ph),
            ]

    def draw_outline(self, color=None, alpha=1/2, **kwargs):
        if self.flip_params is None:
            dx = dy = 0
        elif self.flip_direction is None:
            dummy, rummy, dx, dy = self.flip_params
        else:
            dx, dy = self.flip_direction
        points = self.get_light_points(dx, dy, 0, 0)
        gl.glColor4f(*(color or self.instructions_color), alpha)
        gl.glBegin(gl.GL_LINE_STRIP)
        for point in points:
            gl.glVertex2f(*point)
        if dx == dy == 0:
            gl.glVertex2f(*points[0])
        gl.glEnd()

    def accepts_key(self, key):
        return bool(self.keyhole)

    def unlock(self, key):
        if self.keyhole:
            self.unlocked = key


Layer = Flippable


@attrs(repr=False)
class Sprite(Flippable):
    u = attrib(0)
    v = attrib(0)
    texture_width = attrib()
    texture_height = attrib()

    @texture_width.default
    def _default_w(self):
        return self.width

    @texture_height.default
    def _default_h(self):
        return self.height

    def draw(self, zoom, **kwargs):
        try:
            texture = self.texture
        except AttributeError:
            texture = spritesheet_texture.get_region(
                self.u, self.v, self.texture_width, self.texture_height-1)
            self.texture = texture

        texture.blit(0, 0, width=self.width, height=self.height-1)

        super().draw(zoom=zoom, **kwargs)

    def pixel(self, x, y):
        return spritesheet_data[self.v+y][self.u+x]

    def hit_test_grab(self, x, y):
        return self.pixel(int(x), int(y))

    def hit_test_overlap(self, s, a, b, mode):
        dists = []
        a += 0.0001
        a -= 1
        b -= 0.0001
        for x in range(self.width):
            for y in range(self.height):
                if self.pixel(x, y):
                    if mode == 'x+':
                        if a <= y < b:
                            dists.append(s-x-1)
                    elif mode == 'x-':
                        if a <= y < b:
                            dists.append(x-s)
                    elif mode == 'y+':
                        if a <= x < b:
                            dists.append(s-y-1)
                    elif mode == 'y-':
                        if a <= x < b:
                            dists.append(y-s)
        if not dists:
            return -100
        else:
            return min(dists)



import sys

if sys.version_info < (3, 5):
    # 3.5 shoud be fine, but don't advertise that...
    text = 'You need Python 3.6+ to play this game.'
else:
    text = 'You should not be seeing this...'

try:
    from pyglet.window import Window
    from pyglet.text import Label
    from pyglet.app import run
    from pyglet.clock import schedule

    window = Window(width=800, height=600)
    label = Label(text)
    label2 = Label('This is Python %s.' %
                   '.'.join(str(x) for x in sys.version_info[:3]))

    @window.event
    def on_draw():
        window.clear()
        label.y = window.height / 2 + (label2.content_height or 30)
        label.x = (window.width - (label.content_width or 0)) / 2
        label.draw()
        label2.y = window.height / 2 - (label2.content_height or 30)
        label2.x = (window.width - (label2.content_width or 0)) / 2
        label2.draw()

    @schedule
    def tick(dt):
        pass

    main = run
except:
    raise RuntimeError(text)



import math
from math import pi
import random
import colorsys

from pyglet import gl

from jrti_game.flippable import Flippable, Layer
from jrti_game.text import Letter
from jrti_game.data import sprites, spritesheet_data
from jrti_game.util import reify, draw_rect, clamp
from jrti_game.state import state

tau = pi*2


class Bug(Flippable):
    transparent = True
    homing = False
    homed = False
    s_to_homing = 30

    def __init__(self, **kwargs):
        kwargs.setdefault('width', 15)
        kwargs.setdefault('height', 15)
        kwargs.setdefault('instructions', 'Bug')
        kwargs.setdefault('color', colorsys.hsv_to_rgb(
            random.uniform(0, 1),
            random.uniform(0.2, 0.4),
            random.uniform(0.8, 0.9),
        ))
        self.rotation = random.uniform(-180, 180)
        self.speed = self.min_speed = 190
        self.moment = random.uniform(-20, 20)
        super().__init__(**kwargs)
        self.x = random.uniform(
            0, self.parent.width - self.width * self.scale)
        self.y = random.uniform(
            0, self.parent.height - self.height * self.scale)

    def draw(self, **kwargs):
        gl.glPushMatrix()
        gl.glTranslatef(7.5, 7.5, 0)
        gl.glRotatef(self.rotation, 0, 0, 1)
        sprites['bug'].blit(-5.5, -5.5)

        if self.min_speed > 100:
            leg_time = state.time*32
        else:
            leg_time = state.time*16

        legx = math.sin(leg_time) * 0.5
        legy = 0
        if self.flip_params or state.tool.grabbing is self:
            legx /= 2
            legy = math.cos(leg_time * 8/7) * 0.25
        sprites['bugleg'].blit(-3.5+legx, -5.5+legy)
        gl.glScalef(1, -1, 1)
        sprites['bugleg'].blit(-3.5-legx, -5.5+legy)
        gl.glPopMatrix()

    def tick(self, dt):
        super().tick(dt)
        angle = math.radians(self.rotation)

        ox, oy = self.parent.o_coords
        if (((self.y - oy)/1.5) ** 2 + (self.x - ox) ** 2) < 16**2:
            self.parent.at_o(self)
            if not self.parent:
                return

        hx, hy = self.parent.home_coords
        d = (self.y - hy) ** 2 + (self.x - hx) ** 2
        self.home_distance2 = d
        homed = d < 7**2
        if homed:
            if not self.homed:
                self.homed = True
                self.speed = 1
        if (((self.y - hy) / 1.5) ** 2 + (self.x - hx) ** 2) < 5**2:
            self.min_speed = self.speed = 5
        elif self.homing:
            a = (angle - math.atan2(self.y - hy, self.x - hx) + pi) % tau - pi
            self.moment *= 0.075 ** dt
            if a > tau/180:
                self.moment += math.degrees(a) * dt * 5
            self.rotation += clamp(
                math.degrees(a * dt * 3 * (math.exp(-d/5000))),
                -abs(a), abs(a))

        if not self.flip_params and state.tool.grabbing is not self:
            self.moment *= 0.25 ** dt
            if random.uniform(0, 0.1) < dt:
                self.moment += random.uniform(-60, 60)
            if random.uniform(0, 1) < dt:
                self.moment += random.uniform(-180, 180)
            self.rotation += self.moment * dt
            s = dt * self.speed
            dx = math.cos(angle) * s
            dy = math.sin(angle) * s
            nx = self.x + dx
            ny = self.y + dy
            self.moment += random.uniform(-dt, dt)

            max_x = self.parent.width - self.width*self.scale
            max_y = (self.parent.height - self.height*self.scale)

            if 0 <= nx < max_x and 0 <= ny < max_y:
                self.x = nx
                self.y = ny
            else:
                self.x -= dx
                self.y -= dy

                if not (0 <= self.x < max_x and 0 <= self.y < max_y):
                    self.die()

                self.rotation += 180
                self.moment += random.uniform(-90, 90)

                if abs(self.moment) > 180:
                    self.moment *= 1/2

                ms = self.min_speed
                self.speed += random.uniform(-ms/4, ms/4)
                if abs(self.speed) > ms*2:
                    self.speed *= 1/2
                if self.speed < ms:
                    self.speed = ms
        else:
            if random.uniform(0, 1/10) < dt:
                self.rotation += random.uniform(-5, 5)

        if random.uniform(0, self.s_to_homing) < dt:
            self.homing = True

    def hit_test_grab(self, x, y):
        x -= self.width / 2
        y -= self.height / 2
        s = math.sin(math.radians(self.rotation))
        c = math.cos(math.radians(self.rotation))
        x = x * c - y * s
        y = x * s + y * c
        x += self.width / 2 + 1
        y += self.height / 2 - 5

        if x > 11 or y > 11:
            return False
        return spritesheet_data[28 + int(y)][3*8+1 + int(x)]

    def hit_test_slab(self, x, y, w, h):
        return False


class BugArena(Layer):
    def draw(self, **kwargs):
        gl.glColor3f(.1, .1, .1)
        draw_rect(self.width, self.height)
        super().draw(**kwargs)

    def add_child(self, child):
        super().add_child(child)
        if isinstance(child, Letter):
            child.hides_bugs = True
            child.num_hiding_bugs = 2
            child.bgcolor = self.bgcolor

    def spawn_bug(self, letter=None, s_to_homing=30):
        orig_letter = letter
        while not isinstance(letter, Letter):
            letter = random.choice(self.children)

        closest = 1000
        closest_i = 0
        num_close_bugs = 0
        for i, child in enumerate(self.children):
            dist2 = getattr(child, 'home_distance2', 1000)
            if dist2 < 8**2:
                num_close_bugs += 1
                if dist2 < closest:
                    closest = dist2 + i/10
                    closest_i = i

        if num_close_bugs > 15:
            self.children[closest_i].die()
        bug = Bug(parent=self)
        while True:
            if letter.scale < 8 and random.randrange(0, 6):
                continue
            x = random.randrange(0, letter.width)
            y = random.randrange(0, letter.height)
            if all(letter.pixel(x+xx, y+yy)
                    for xx in (0, 1) for yy in (0, 1)):
                bug.x = letter.x + x * letter.scale
                bug.y = letter.y + y * letter.scale
                if orig_letter is None:
                    bug.rotation = random.normalvariate(90, 30)
                else:
                    bug.rotation = random.uniform(0, 180)
                if random.randrange(2):
                    bug.rotation = -bug.rotation
                if letter is self.o_letter:
                    bug.x += math.sin(math.radians(bug.rotation)) * 16
                    bug.y += math.sin(math.radians(bug.rotation)) * 24
                bug.s_to_homing = s_to_homing
                self.children.sort(key=lambda c: -isinstance(c, Bug))
                return bug

    def tick(self, dt):
        super().tick(dt)
        num_bugs = len([c for c in self.children
                        if isinstance(c, Bug) and not c.homed])
        s_to_bug = clamp((num_bugs-1/2)*4, 1, 100)
        if random.uniform(0, s_to_bug) < dt:
            self.spawn_bug()

    @property
    def o_coords(self):
        l = self.o_letter
        return l.x + l.scale * 2.5, l.y + l.scale * 3

    @reify
    def home_coords(self):
        return self.o_coords

    @reify
    def o_letter(self):
        for child in self.children:
            if child.instructions == 'O':
                child.num_hiding_bugs += 10
                return child

    def hit_test_overlap(self, s, a, b, mode):
        a += 0.0001
        a -= 1
        b -= 0.0001
        if mode == 'x+':
            if a < self.height and 0 < b:
                return s-self.width-1
        elif mode == 'x-':
            if a < self.height and 0 <= b:
                return -s
        elif mode == 'y+':
            if a < self.width and 0 <= b:
                return s-self.height-1
        elif mode == 'y-':
            if a < self.width and 0 <= b:
                return -s
        return -100

    def at_o(self, bug):
        bug.die()
        self.o_letter.num_hiding_bugs = clamp(
            self.o_letter.num_hiding_bugs + 1, 0, 20)



import sys
import traceback

from jrti_game.data import load_instructions, default_keymap
from jrti_game.state import state
from jrti_game.key_names import key_symbols


instruction_texts = load_instructions()


def get_instruction_text(key):
    text = instruction_texts.get(key, instruction_texts['None'])
    return text.format_map(substitutions)


class Substitutions(dict):
    def __missing__(self, name):
        try:
            if name == 'state':
                return state
            if name == 'end_time':
                if state.end_time is None:
                    return 'some cheaty way'
                else:
                    return format_end_time(state.end_time)
            action = default_keymap.get(name)
            if action:
                for current_key, current_action in state.keymap.items():
                    if current_action == action:
                        try:
                            return key_symbols[current_key][1]
                        except KeyError:
                            return 'Key {0:x}'.format(current_key)
                return '?'

            raise LookupError()
        except:
            traceback.print_exc()
            print(file=sys.stderr)
            raise

substitutions = Substitutions()


def format_end_time(time):
    def s(val, text):
        return "{val} {text}{s}".format(val=val, text=text,
                                        s='' if val==1 else 's')

    time = int(time)

    if time == 0:
        return 'some cheaty way'

    result = []
    for number, unit in (
            (60, 'second'),
            (60, 'minute'),
            (24, 'hour'),
            (7, 'day'),
    ):
        time, val = divmod(time, number)
        if val:
            result.append(s(val, unit))
    if time:
        result.append(s(time, 'week'))

    result.reverse()
    result[-2:] = [' and '.join(result[-2:])]
    return ', '.join(result)


def test_eq(a, b):
    if a != b:
        print('Fail:')
        print('    got:', repr(a))
        print('    exp:', repr(b))

test_eq(format_end_time(0), 'some cheaty way')
test_eq(format_end_time(1), '1 second')
test_eq(format_end_time(2), '2 seconds')
test_eq(format_end_time(60), '1 minute')
test_eq(format_end_time(62), '1 minute and 2 seconds')
test_eq(format_end_time(3600), '1 hour')
test_eq(format_end_time(3663), '1 hour, 1 minute and 3 seconds')
test_eq(format_end_time(3600*24), '1 day')
test_eq(format_end_time(3601*24), '1 day and 24 seconds')
test_eq(format_end_time(3600*24*7), '1 week')
test_eq(format_end_time(3600*24*7*2), '2 weeks')
test_eq(format_end_time(3600*24*7*2+3600*24*3+3600*7+60*2+3),
        '2 weeks, 3 days, 7 hours, 2 minutes and 3 seconds')
test_eq(format_end_time(3600*24*7*200), '200 weeks')



from jrti_game.state import state


def mouse_to_logical(mx, my):
    cx, cy = state.center
    lx = (mx - 400) / state.zoom + cx
    ly = (my - 300) / state.zoom + cy
    return lx, ly


def logical_to_mouse(lx, ly):
    cx, cy = state.center
    mx = (lx - cx) * state.zoom + 400
    my = (ly - cy) * state.zoom + 300

    return mx, my


def mouse_to_main_screen(mx, my):
    lx, ly = mouse_to_logical(mx, my)
    return lx - state.main_screen.x, ly - state.main_screen.y


def tool_to_main_screen(tx, ty):
    sx, sy = mouse_to_logical(tx + 400, ty + 300)
    return sx - state.main_screen.x, sy - state.main_screen.y


def tool_to_logical(tx, ty):
    sx, sy = mouse_to_logical(tx + 400, ty + 300)
    return sx, sy



# This file regenerates itself

from pyglet.window import key


if __name__ == '__main__':
    with open(__file__) as f:
        for line in f:
            print(line, end='')
            if line.strip() == '####':
                break

    NAMES = {
        # Based in part on:
        # https://wincent.com/wiki/Unicode_representations_of_modifier_keys
        'SPACE': '␣',
        'ENTER': '⏎',
        'RETURN': '⏎',
        'BACKSPACE': '⌫',
        'DELETE': '⌦',
        'TAB': '⇥',
        'ESCAPE': '×',
        'PAUSE': '⏸',
        'INSERT': 'Ｉ',
        'MENU': '☰',
        'HELP': '‽',
        'CAPSLOCK': '⇪',
        'DIVIDE': '÷',

        'CTRL': '⌃',
        'SHIFT': '⇧',
        'COMMAND': '⌘',
        'OPTION': '⌥',
        'WINDOWS': '⊞',
        'META': '⌃',
        'ALT': 'Ａ',

        'ADD': '+',
        'SUBTRACT': '-',
        'MULTIPLY': '×',
        'EQUAL': '=',
        'SEPARATOR': ' ',
        'DECIMAL': ',',

        'HOME': '⇱',
        'END': '⇲',
        'PAGEUP': '⇑',
        'PAGEDOWN': '⇓',
        'PAGE_UP': '⇑',
        'PAGE_DOWN': '⇓',

        'LEFT': '←',
        'UP': '↑',
        'DOWN': '↓',
        'RIGHT': '→',
    }

    LABELS = {
        **NAMES,
        'SPACE': 'Space',
        'ENTER': 'Enter',
        'RETURN': 'Return',
        'BACKSPACE': 'Backspace',
        'DELETE': 'Delete',
        'TAB': 'Tab',
        'ESCAPE': 'Esc',
        'PAUSE': 'Pause',
        'INSERT': 'Insert',
        'MENU': 'Menu',
        'HELP': 'Help',
        'CAPSLOCK': 'Caps Lock',

        'CTRL': 'Ctrl',
        'SHIFT': 'Shift',
        'COMMAND': 'Command',
        'OPTION': 'Option',
        'WINDOWS': 'Windows',
        'ALT': 'Alt',

        'SEPARATOR': 'Separator',

        'HOME': 'Home',
        'END': 'End',
        'PAGEUP': 'Page Up',
        'PAGEDOWN': 'Page Down',
        'PAGE_UP': 'Page Up',
        'PAGE_DOWN': 'Page Down',
        'SCROLLLOCK': 'Scroll Lock',
    }

    print('key_symbols = {')
    symbols = [k for k in dir(key) if k.isupper()]
    symbols.extend('_{}'.format(n) for n in range(10))
    for name in symbols:
        if name.startswith(('MOTION_', 'MOD_')):
            continue
        value = getattr(key, name)
        note = ''
        if name in NAMES:
            symbol = NAMES[name]
            label = LABELS[name]
        elif name.startswith('L') and name[1:] in NAMES:
            symbol = NAMES[name[1:]]
            label = 'Left ' + LABELS[name[1:]]
        elif name.startswith('R') and name[1:] in NAMES:
            symbol = NAMES[name[1:]]
            label = 'Right ' + LABELS[name[1:]]
        elif name.startswith('NUM_') and name[4:] in NAMES:
            symbol = NAMES[name[4:]]
            label = 'Numpad ' + LABELS[name[4:]]
        elif name.startswith('NUM_') and name[4:].isnumeric():
            symbol = 'n ' + name[4:]
            label = 'Numpad ' + name[4:]
        elif name.startswith('F') and name[1:].isnumeric():
            symbol = label = name
        elif len(name) == 1:
            symbol = label = name
        elif ord(' ') < value <= ord('~'):
            symbol = label = chr(value)
        else:
            if name.startswith('NUM_'):
                symbol = 'n' + name[4:6].lower()
                label = name
            else:
                symbol = name[:3].lower()
                label = name.title()
            note = '            # ? ' + name.lower()
        if name in LABELS:
            label = LABELS[name]
        print('    key.{}: {},{}'.format(
                name,
                repr((symbol, label)),
                note,
            ))
    print('}')

####
key_symbols = {
    key.A: ('A', 'A'),
    key.AMPERSAND: ('&', '&'),
    key.APOSTROPHE: ("'", "'"),
    key.ASCIICIRCUM: ('^', '^'),
    key.ASCIITILDE: ('~', '~'),
    key.ASTERISK: ('*', '*'),
    key.AT: ('@', '@'),
    key.B: ('B', 'B'),
    key.BACKSLASH: ('\\', '\\'),
    key.BACKSPACE: ('⌫', 'Backspace'),
    key.BAR: ('|', '|'),
    key.BEGIN: ('beg', 'Begin'),            # ? begin
    key.BRACELEFT: ('{', '{'),
    key.BRACERIGHT: ('}', '}'),
    key.BRACKETLEFT: ('[', '['),
    key.BRACKETRIGHT: (']', ']'),
    key.BREAK: ('bre', 'Break'),            # ? break
    key.C: ('C', 'C'),
    key.CANCEL: ('can', 'Cancel'),            # ? cancel
    key.CAPSLOCK: ('⇪', 'Caps Lock'),
    key.CLEAR: ('cle', 'Clear'),            # ? clear
    key.COLON: (':', ':'),
    key.COMMA: (',', ','),
    key.D: ('D', 'D'),
    key.DELETE: ('⌦', 'Delete'),
    key.DOLLAR: ('$', '$'),
    key.DOUBLEQUOTE: ('"', '"'),
    key.DOWN: ('↓', '↓'),
    key.E: ('E', 'E'),
    key.END: ('⇲', 'End'),
    key.ENTER: ('⏎', 'Enter'),
    key.EQUAL: ('=', '='),
    key.ESCAPE: ('×', 'Esc'),
    key.EXCLAMATION: ('!', '!'),
    key.EXECUTE: ('exe', 'Execute'),            # ? execute
    key.F: ('F', 'F'),
    key.F1: ('F1', 'F1'),
    key.F10: ('F10', 'F10'),
    key.F11: ('F11', 'F11'),
    key.F12: ('F12', 'F12'),
    key.F13: ('F13', 'F13'),
    key.F14: ('F14', 'F14'),
    key.F15: ('F15', 'F15'),
    key.F16: ('F16', 'F16'),
    key.F17: ('F17', 'F17'),
    key.F18: ('F18', 'F18'),
    key.F19: ('F19', 'F19'),
    key.F2: ('F2', 'F2'),
    key.F20: ('F20', 'F20'),
    key.F3: ('F3', 'F3'),
    key.F4: ('F4', 'F4'),
    key.F5: ('F5', 'F5'),
    key.F6: ('F6', 'F6'),
    key.F7: ('F7', 'F7'),
    key.F8: ('F8', 'F8'),
    key.F9: ('F9', 'F9'),
    key.FIND: ('fin', 'Find'),            # ? find
    key.FUNCTION: ('fun', 'Function'),            # ? function
    key.G: ('G', 'G'),
    key.GRAVE: ('`', '`'),
    key.GREATER: ('>', '>'),
    key.H: ('H', 'H'),
    key.HASH: ('#', '#'),
    key.HELP: ('‽', 'Help'),
    key.HOME: ('⇱', 'Home'),
    key.I: ('I', 'I'),
    key.INSERT: ('Ｉ', 'Insert'),
    key.J: ('J', 'J'),
    key.K: ('K', 'K'),
    key.L: ('L', 'L'),
    key.LALT: ('Ａ', 'Left Alt'),
    key.LCOMMAND: ('⌘', 'Left Command'),
    key.LCTRL: ('⌃', 'Left Ctrl'),
    key.LEFT: ('←', '←'),
    key.LESS: ('<', '<'),
    key.LINEFEED: ('lin', 'Linefeed'),            # ? linefeed
    key.LMETA: ('⌃', 'Left ⌃'),
    key.LOPTION: ('⌥', 'Left Option'),
    key.LSHIFT: ('⇧', 'Left Shift'),
    key.LWINDOWS: ('⊞', 'Left Windows'),
    key.M: ('M', 'M'),
    key.MENU: ('☰', 'Menu'),
    key.MINUS: ('-', '-'),
    key.MODESWITCH: ('mod', 'Modeswitch'),            # ? modeswitch
    key.N: ('N', 'N'),
    key.NUMLOCK: ('num', 'Numlock'),            # ? numlock
    key.NUM_0: ('n 0', 'Numpad 0'),
    key.NUM_1: ('n 1', 'Numpad 1'),
    key.NUM_2: ('n 2', 'Numpad 2'),
    key.NUM_3: ('n 3', 'Numpad 3'),
    key.NUM_4: ('n 4', 'Numpad 4'),
    key.NUM_5: ('n 5', 'Numpad 5'),
    key.NUM_6: ('n 6', 'Numpad 6'),
    key.NUM_7: ('n 7', 'Numpad 7'),
    key.NUM_8: ('n 8', 'Numpad 8'),
    key.NUM_9: ('n 9', 'Numpad 9'),
    key.NUM_ADD: ('+', 'Numpad +'),
    key.NUM_BEGIN: ('nbe', 'NUM_BEGIN'),            # ? num_begin
    key.NUM_DECIMAL: (',', 'Numpad ,'),
    key.NUM_DELETE: ('⌦', 'Numpad Delete'),
    key.NUM_DIVIDE: ('÷', 'Numpad ÷'),
    key.NUM_DOWN: ('↓', 'Numpad ↓'),
    key.NUM_END: ('⇲', 'Numpad End'),
    key.NUM_ENTER: ('⏎', 'Numpad Enter'),
    key.NUM_EQUAL: ('=', 'Numpad ='),
    key.NUM_F1: ('nf1', 'NUM_F1'),            # ? num_f1
    key.NUM_F2: ('nf2', 'NUM_F2'),            # ? num_f2
    key.NUM_F3: ('nf3', 'NUM_F3'),            # ? num_f3
    key.NUM_F4: ('nf4', 'NUM_F4'),            # ? num_f4
    key.NUM_HOME: ('⇱', 'Numpad Home'),
    key.NUM_INSERT: ('Ｉ', 'Numpad Insert'),
    key.NUM_LEFT: ('←', 'Numpad ←'),
    key.NUM_MULTIPLY: ('×', 'Numpad ×'),
    key.NUM_NEXT: ('nne', 'NUM_NEXT'),            # ? num_next
    key.NUM_PAGE_DOWN: ('⇓', 'Numpad Page Down'),
    key.NUM_PAGE_UP: ('⇑', 'Numpad Page Up'),
    key.NUM_PRIOR: ('npr', 'NUM_PRIOR'),            # ? num_prior
    key.NUM_RIGHT: ('→', 'Numpad →'),
    key.NUM_SEPARATOR: ('\xa0', 'Numpad Separator'),
    key.NUM_SPACE: ('␣', 'Numpad Space'),
    key.NUM_SUBTRACT: ('-', 'Numpad -'),
    key.NUM_TAB: ('⇥', 'Numpad Tab'),
    key.NUM_UP: ('↑', 'Numpad ↑'),
    key.O: ('O', 'O'),
    key.P: ('P', 'P'),
    key.PAGEDOWN: ('⇓', 'Page Down'),
    key.PAGEUP: ('⇑', 'Page Up'),
    key.PARENLEFT: ('(', '('),
    key.PARENRIGHT: (')', ')'),
    key.PAUSE: ('⏸', 'Pause'),
    key.PERCENT: ('%', '%'),
    key.PERIOD: ('.', '.'),
    key.PLUS: ('+', '+'),
    key.POUND: ('#', '#'),
    key.PRINT: ('pri', 'Print'),            # ? print
    key.Q: ('Q', 'Q'),
    key.QUESTION: ('?', '?'),
    key.QUOTELEFT: ('`', '`'),
    key.R: ('R', 'R'),
    key.RALT: ('Ａ', 'Right Alt'),
    key.RCOMMAND: ('⌘', 'Right Command'),
    key.RCTRL: ('⌃', 'Right Ctrl'),
    key.REDO: ('red', 'Redo'),            # ? redo
    key.RETURN: ('⏎', 'Return'),
    key.RIGHT: ('→', '→'),
    key.RMETA: ('⌃', 'Right ⌃'),
    key.ROPTION: ('⌥', 'Right Option'),
    key.RSHIFT: ('⇧', 'Right Shift'),
    key.RWINDOWS: ('⊞', 'Right Windows'),
    key.S: ('S', 'S'),
    key.SCRIPTSWITCH: ('scr', 'Scriptswitch'),            # ? scriptswitch
    key.SCROLLLOCK: ('scr', 'Scroll Lock'),            # ? scrolllock
    key.SELECT: ('sel', 'Select'),            # ? select
    key.SEMICOLON: (';', ';'),
    key.SLASH: ('/', '/'),
    key.SPACE: ('␣', 'Space'),
    key.SYSREQ: ('sys', 'Sysreq'),            # ? sysreq
    key.T: ('T', 'T'),
    key.TAB: ('⇥', 'Tab'),
    key.U: ('U', 'U'),
    key.UNDERSCORE: ('_', '_'),
    key.UNDO: ('und', 'Undo'),            # ? undo
    key.UP: ('↑', '↑'),
    key.V: ('V', 'V'),
    key.W: ('W', 'W'),
    key.X: ('X', 'X'),
    key.Y: ('Y', 'Y'),
    key.Z: ('Z', 'Z'),
    key._0: ('0', '0'),
    key._1: ('1', '1'),
    key._2: ('2', '2'),
    key._3: ('3', '3'),
    key._4: ('4', '4'),
    key._5: ('5', '5'),
    key._6: ('6', '6'),
    key._7: ('7', '7'),
    key._8: ('8', '8'),
    key._9: ('9', '9'),
}



import math

from pyglet import gl

from jrti_game.data import sprites
from jrti_game.state import state
from jrti_game.util import clamp, interp
from jrti_game.coords import tool_to_main_screen, tool_to_logical


class Tool:
    size = 0
    scale = 0
    speed = 0
    name = 'none'
    grabbing = None

    def draw(self):
        pass

    def deactivate(self):
        pass

    def tick(self, dt):
        pass

    def move(self, dx, dy):
        state.tool_x += dx
        state.tool_y += dy
        self.speed = dx + dy


class Grabby(Tool):
    size = 8*2-2
    scale = 3
    active = False
    grab_time = 0
    name = 'grabby'
    grab_duration = 1/8

    def draw(self):
        gt = clamp((state.time - self.grab_time) / self.grab_duration, 0, 1)
        if not self.active:
            gt = 1-gt

        gl.glRotatef(90 + self.speed/2 + gt * 45, 0, 0, 1)

        gl.glColor3f(
            interp(0.9, 0.1, gt**(2)),
            interp(0.1, 0.9, gt**(1/2)),
            0.4 + 0.06 * 4 * gt * (1-gt)
        )
        for i in range(4):
            sprites['grabby'].blit(0, 0)
            gl.glRotatef(90, 0, 0, 1)

        gl.glColor3f(interp(0.9 * abs(math.cos(state.time * 3)), 0.1, gt),
                     interp(0.1 * math.sin(state.time * 3)**2, 0.9, gt),
                     0.2 * math.sin(state.time * 3)**2)
        tr = interp(
            (abs(math.sin(state.time * 3)) * 2 - 1),
            (-4.5),
            gt,
        )
        for i in range(4):
            sprites['grabtooth'].blit(tr, tr)
            gl.glRotatef(90, 0, 0, 1)

    def deactivate(self):
        if self.active:
            self.active = False
            self.grabbing = False
            if self.grab_time < state.time - self.grab_duration:
                self.grab_time = state.time

    def grab(self):
        self.active = not self.active
        if not self.active:
            self.grabbing = False
        if self.grab_time < state.time - self.grab_duration:
            self.grab_time = state.time

    def tick(self, dt):
        if self.grab_time < state.time - self.grab_duration:
            if self.active and not self.grabbing:
                x, y = tool_to_main_screen(state.tool_x, state.tool_y)
                for obj, x, y in state.main_screen.hit_test_all(x, y):
                    if obj.hit_test_grab(x, y):
                        if obj.unlocked:
                            obj.unlocked.anim_start = state.time
                            continue
                        self.grabbing = obj
                        break
                else:
                    self.deactivate()

    def move(self, dx, dy):
        if self.grabbing:
            scale = state.zoom
            g = self.grabbing
            obj = g.parent
            while obj:
                scale *= obj.scale
                obj = obj.parent

        if self.grabbing and self.grabbing.grab_move(dx/scale, dy/scale):
            pass
        elif self.grabbing:
            gdx = dx
            gdy = dy

            # Find appropriate distances to all siblings
            # - If negative - overlapping, don't care
            # - If distance is less than dN, limit dN to that
            e0 = -0.0001
            for sibling in g.parent.children:
                x = (g.x - sibling.x) / sibling.scale
                y = (g.y - sibling.y) / sibling.scale
                w = g.width * g.scale / sibling.scale
                h = g.height * g.scale / sibling.scale
                if gdx < 0:
                    dist = sibling.hit_test_overlap(x, y, y+h, 'x+')
                    dist *= sibling.scale
                    if e0 <= dist <= -gdx:
                        gdx = -dist
                if gdx > 0:
                    dist = sibling.hit_test_overlap(x+w, y, y+h, 'x-')
                    dist *= sibling.scale
                    if e0 <= dist <= gdx:
                        gdx = dist
                if gdy < 0:
                    dist = sibling.hit_test_overlap(y, x, x+w, 'y+')
                    dist *= sibling.scale
                    if e0 <= dist <= -gdy:
                        gdy = -dist
                if gdy > 0:
                    dist = sibling.hit_test_overlap(y+h, x, x+w, 'y-')
                    dist *= sibling.scale
                    if e0 <= dist <= gdy:
                        gdy = dist

            g.x = clamp(g.x + gdx / scale, 0,
                        g.parent.width - g.width * g.scale)
            g.y = clamp(g.y + gdy / scale,
                        0, g.parent.height - g.height * g.scale)

            if gdx or gdy and g.hides_bugs:
                if (state.time - g.last_moved) > 0.5:
                    g.num_hiding_bugs += (state.time - g.last_moved) ** (1/3)
                g.last_moved = state.time

                hb = g.num_hiding_bugs
                db = abs(gdx + gdy) / self.scale
                left = clamp(hb - db, 0, 10)
                for i in range(int(left), int(hb)):
                    g.parent.spawn_bug(g, s_to_homing=4)
                g.num_hiding_bugs = left

        super().move(dx, dy)
        x, y = tool_to_main_screen(state.tool_x, state.tool_y)
        for obj, x, y in state.main_screen.hit_test_all(x, y):
            if obj is self.grabbing and obj.hit_test(x, y):
                break
        else:
            self.deactivate()


class Key(Tool):
    size = 7
    scale = 3
    anim_start = 0
    name = 'key'
    locked = None
    locking = False
    unlocked = False
    lock = None
    visible_len = 19
    fully_inserted = False
    lock_zoom = 0
    open_anim_start = 0

    def draw(self):
        anim = clamp((state.time - self.anim_start) * 4, 0, 1)
        length = self.visible_len
        if not self.locking:
            anim = 1-anim
        if self.unlocked:
            anim = 1
        gl.glRotatef(anim * 180, 1, 0, 0)
        if self.fully_inserted:
            gl.glColor3f(0.1, 0.9, 0.1 + anim*0.9)
        else:
            t = (self.visible_len-9) / 10
            gl.glColor3f(0.9, 0.9 - t*0.8, 0.1)
        texture = sprites['key']
        if length < 15:
            texture = texture.get_region(0, 0, width=min(length, 15), height=8)
        texture.blit(-4.5, -4.5)
        for i, num in enumerate(state.key_config):
            if i + 15 > length:
                break
            if i + 16 > length:
                width = length % 1
            else:
                width = 1
            if num < 5:
                y = -0.75+(num-5)/2
                height = 1-(num-5)/2+.25
            else:
                y = -0.75
                height = 1+(num-5)/2
            sprites['square'].blit(x=15-4.5+i, y=y, width=width, height=height)

    def turn(self):
        if self.anim_start < state.time - 1/4:
            self.locking = not self.locked
            self.anim_start = state.time

    def tick(self, dt):
        if self.anim_start:
            t = state.time - self.anim_start
            if self.locking:
                accepts = not self.lock or self.lock.accepts_key(self)
                if t > 1/4 and accepts:
                    if self.lock and self.fully_inserted:
                        self.lock.unlock(self)
                        self.unlocked = True
                        self.anim_start = self.open_anim_start = state.time
                        state.tool = Tool()
                    else:
                        self.locking = False
                        self.anim_start = state.time
                elif t > 1/16 and (not accepts or not self.fully_inserted):
                    if self.lock and self.visible_len < 19:
                        self.locking = False
                        self.anim_start = state.time - (1/4-t)

    def deactivate(self):
        self.lock = None
        self.visible_len = 19
        self.fully_inserted = False

    def move(self, dx, dy):
        if self.unlocked:
            return

        if dx < 0 or dy:
            self.deactivate()

        if dx > 0:
            if not self.lock:
                x, y = tool_to_main_screen(state.tool_x+14.5*3, state.tool_y)
                for obj, x, y in state.main_screen.hit_test_all(x, y):
                    if obj.keyhole and not obj.unlocked:
                        zoom = state.zoom
                        par = obj
                        while par:
                            zoom *= par.scale
                            par = par.parent
                        self.lock_zoom = zoom
                        kx, ky, kw, kh = obj.keyhole
                        kx -= dx / zoom
                        kw += dx / zoom
                        kmin, kmax = obj.keyhole_params
                        if (kx < x < kx+kw and ky < y < ky+kh and
                                kmin < kh * zoom < kmax):
                            self.lock = obj
                            break
            if self.lock:
                kx, ky, kw, kh = self.lock.keyhole
                x, y = tool_to_logical(state.tool_x+4.5*3, state.tool_y)
                parents = []
                obj = self.lock
                while obj:
                    parents.append(obj)
                    obj = obj.parent
                for obj in reversed(parents):
                    x = (x - obj.x) / obj.scale
                    y = (y - obj.y) / obj.scale

                insert_remaining = (kx+kw - x)*self.lock_zoom
                if dx > insert_remaining:
                    dx = insert_remaining
                    self.fully_inserted = True
                    self.obj_pos = x, y, self.lock_zoom
                elif y < ky+kh/2:
                    dy += min(dx, abs(ky+kh/2 - y)*self.lock_zoom)
                elif y > ky+kh/2:
                    dy -= min(dx, abs(ky+kh/2 - y)*self.lock_zoom)
                insert_remaining = (kx+kw - x)*self.lock_zoom - dx
                self.visible_len = 9 + insert_remaining / 3

        super().move(dx, dy)

    def draw_stuck(self):
        gl.glPushMatrix()
        x, y, z = self.obj_pos
        gl.glTranslatef(x, y, 0)
        gl.glScalef(3/z, 3/z, 1)
        gl.glColor3f(
            clamp((state.time - self.anim_start)*2, 0.1, 1),
            clamp((state.time - self.anim_start)*2, 0.9, 1),
            clamp((state.time - self.anim_start)*2, 0.9, 1),
        )
        sprites['key_head'].blit(-9, -4.5)
        gl.glPopMatrix()



from pyglet import gl

from jrti_game.flippable import Flippable
from jrti_game.data import sprites, spritesheet_data
from jrti_game.util import draw_rect_xy
from jrti_game.state import state
from jrti_game.text import draw_string
from jrti_game.util import clamp


class Adjuster(Flippable):
    endgen = None

    def draw(self, zoom, **kwargs):
        gl.glColor3f(1, 1, 1)
        if zoom > 1:
            draw_rect_xy(11, 12, 1, 80)
        elif zoom > 1/64:
            gl.glBegin(gl.GL_LINES)
            gl.glVertex2f(12, 12)
            gl.glVertex2f(12, 12+80)
            gl.glEnd()
        if zoom > 1/64:
            pos = state.key_config[self.index]
            sprites['adjuster'].blit(0, pos*8)

    def hit_test(self, x, y):
        pos = state.key_config[self.index]
        return 0 < x < 24 and 0 < y-pos*8 < 24

    def hit_test_grab(self, x, y):
        pos = state.key_config[self.index]
        return spritesheet_data[3*8+1+int(y-pos*8)][96+int(x)]

    def grab_move(self, dx, dy):
        pos = state.key_config[self.index]
        pos = clamp(pos + dy*16, 0, 9)
        state.key_config[self.index] = pos
        return True

    def draw_grab(self, *args, **kwargs):
        pass



from pyglet import gl

from jrti_game.flippable import Sprite
from jrti_game.data import sprites, spritesheet_data
from jrti_game.util import draw_rect_xy
from jrti_game.state import state
from jrti_game.text import draw_string
from jrti_game.util import clamp


class Keyhole(Sprite):
    keyhole = 2, 18, 6.5, 13
    keyhole_params = 10, 800
    endgen = None

    def unlock(self, key):
        super().unlock(key)
        self.endgen = self.end_animation()
        next(self.endgen)
        state.ending = True
        state.end_time = state.time
        state.main_screen.update_instructions()
        print('Done! Time = ', state.end_time)

    def tick(self, dt):
        super().tick(dt)
        if self.endgen:
            try:
                self.endgen.send(dt)
            except StopIteration:
                self.endgen = False
                state.ending = False


    def end_animation(self):
        yield
        state.min_zoom = 0.5
        state.zoom_easing = 0.7
        prev_zoom = state.target_zoom

        state.target_zoom = 1
        while state.zoom != state.target_zoom:
            yield

        if prev_zoom != 1:
            target = state.time + 0.25
            while state.time < target:
                yield

        state.target_zoom = 0.5
        while state.zoom != state.target_zoom:
            yield

        state.zoom_easing = 1
        angle = 0
        speed = 0
        state.main_screen.flip_locked = True
        while angle < 150:
            angle = clamp(angle+speed, 0, 150)
            speed += (yield) * 5
            state.main_screen.flip_params = 0, 0, angle, 0

    def accepts_key(self, key):
        return all(int(a) == int(b)
                   for a, b
                   in zip(state.key_config, state.code))



from setuptools import setup

setup(
    name='jrti_game',
    version='1.0.3',
    description='Just Read the Instructions, the game',
    author='Petr Viktorin',
    author_email='encukou@gmail.com',
    license='MIT',
    url='https://pyweek.org/e/instructions/',
    packages=['jrti_game'],
    install_requires=[
        'pyglet == 1.2.4',
        'pypng == 0.0.18',
        'attrs == 17.2.0',
    ],
    entry_points={
        'console_scripts': [
            'jrti-game = jrti_game:main',
        ],
    },
    package_data={'jrti_game': ['data/*', 'data/*/*']},
)



import functools
import contextlib

from pyglet import gl

from jrti_game.data import spritesheet_data, spritesheet_texture, sprites
from jrti_game.flippable import Sprite
from jrti_game.key_names import key_symbols
from jrti_game.util import draw_rect, draw_rect_xy, clamp, draw_keyhole
from jrti_game.state import state


@functools.lru_cache()
def letter_uvwh(character):
    number = specials.get(character, ord(character))
    if number < ord(' '):
        number = 127
    if number > 0xff:
        number = 127
    u = 8 * ((number - ord(' ')) % 32)
    v = 8 * ((number - ord(' ')) // 32)
    for width in range(8, 0, -1):
        if any(spritesheet_data[v+7-i][u+width-1]
               for i in range(8)):
            break
    return u, v, width+1, 9


def text_width(string):
    width = 0
    last_char = None
    for character in string:
        u, v, w, h = letter_uvwh(character)
        width += w + kerns.get((last_char, character), 0)
    return width


specials = {
    '␣': 0x90,
    '⏎': 0x91,
    '⌫': 0x92,
    '⌦': 0x93,
    '⇥': 0x94,
    '×': 0x95,
    '⏸': 0x96,
    'Ｉ': 0x97,
    '☰': 0x98,
    '‽': 0x99,
    '⇪': 0x9a,

    '⌃': 0x9b,
    '⇧': 0x9c,
    '⌘': 0x9d,
    '⌥': 0x9e,
    '⊞': 0x9f,
    'Ａ': 0xb0,

    '⇱': 0xb1,
    '⇲': 0xb2,
    '⇑': 0xb3,
    '⇓': 0xb4,

    '←': 0xb5,
    '↑': 0xb6,
    '↓': 0xb7,
    '→': 0xb8,

    '⌀': 0xb9,
    '÷': 0xba
}

kerns = {
    ('s', 't'): -1,
    ('/', '/'): -1,
}


def draw_string(string, center=False, hcenter=False, x=0, y=0):
    if center:
        x -= (text_width(string)-1) / 2
    if hcenter:
        y -= 4
    for char in string:
        u, v, w, h = letter_uvwh(char)
        if char != ' ':
            spritesheet_texture.get_region(u, v, w, h).blit(x, y)
        x += w


def draw_key(key, x=0, y=0):
    if key in key_symbols:
        txt = key_symbols[key][0]
        draw_string(txt, center=True, hcenter=True, x=x, y=y)
    else:
        gl.glPushMatrix()
        gl.glScalef(1/4, 1/4, 1)
        txt = '{0:016x}'.format(key).replace('0', '⌀')
        draw_string(txt[-16:-12], center=True, hcenter=True, x=x, y=y+12)
        draw_string(txt[-12:-8], center=True, hcenter=True, x=x, y=y+4)
        draw_string(txt[-8:-4], center=True, hcenter=True, x=x, y=y-4)
        draw_string(txt[-4:], center=True, hcenter=True, x=x, y=y+-12)
        gl.glPopMatrix()


class Letter(Sprite):
    def __init__(self, letter, strim=0, etrim=0, **kwargs):
        u, v, width, height = letter_uvwh(letter)
        width += 1-etrim-strim
        kwargs.update(
            texture_width=width,
            texture_height=height,
            u=u-1+strim,
            v=v,
        )
        if 'width' not in kwargs:
            kwargs['width'] = (kwargs['height'] // 9 * width) or 1
        kwargs.setdefault('instructions', letter.upper())
        kwargs.setdefault('margin', 1/2)
        super().__init__(**kwargs)

    def draw(self, zoom, **kwargs):
        super().draw(zoom=zoom, **kwargs)
        if self.keyhole and zoom > 9:
            gl.glColor4f(*self.bgcolor)
            draw_keyhole(self.keyhole)


def letters(string, x=0, y=0, scale=1, center=False, parent=None,
            replace=None):
    if center:
        x -= scale * (text_width(string) // 2)
    starting_x = x
    next_kern = 0
    for character, next_char in zip(string, string[1:] + ' '):
        instructions = character.upper()
        if character == '\n':
            y -= scale * 8
            x = starting_x
        else:
            x += next_kern
            if next_kern < 0:
                strim = 1
                x += scale
            else:
                strim = 0
            next_kern = kerns.get((character, next_char), 0) * scale
            if next_kern < 0:
                etrim = 1
            else:
                etrim = 0
            if character == ' ':
                x += (4 + etrim) * scale
            else:
                cls = SPECIAL_LETTER_CLASSES.get(character, Letter)
                letter = cls(character, x=x, y=y, height=9, scale=scale,
                             strim=strim, etrim=etrim, parent=parent,
                             instructions=instructions)
                x += (letter.width + etrim) * scale - scale
                yield letter


class Oh(Letter):
    def draw(self, **kwargs):
        super().draw(**kwargs)
        gl.glPushMatrix()
        gl.glColor4f(*self.bgcolor)
        gl.glTranslatef(3, 2, 0)
        draw_rect(1, 4)
        gl.glPopMatrix()


@contextlib.contextmanager
def draw_opening(self):
    gl.glPushMatrix()
    draw_rect_xy(*self.innards)
    gl.glColor3f(1, 1, 1)
    draw_rect_xy(*self.innards, mode=gl.GL_LINE_STRIP, last=False)
    gl.glPushMatrix()
    yield
    gl.glPopMatrix()
    right = self.innards[0] + self.innards[2]
    gl.glTranslatef(right, 0, 0)
    t = clamp((state.time - self.unlocked.open_anim_start)*2, 0, 1)**3
    gl.glRotatef(-t*100, 0, 1, 0)
    gl.glTranslatef(-right, 0, 0)
    gl.glColor3f(1, 1, 1)
    draw_rect_xy(*self.innards)
    gl.glColor4f(0, 0, 0, 1/2)
    draw_rect_xy(*self.innards, mode=gl.GL_LINE_STRIP, last=False)
    gl.glPopMatrix()


class Eye(Letter):
    keyhole = 2-1/16, 7, 1/16, 1
    innards = 2+1/8, 7+1/16, 2-3/16, 1-2/16

    def draw(self, zoom, **kwargs):
        super().draw(zoom=zoom, children=False, **kwargs)
        if self.unlocked:
            if zoom > 4:
                gl.glColor3f(0, 0, 0)
            else:
                gl.glColor3f(1, 1, 1)
            with draw_opening(self):
                if zoom > 4:
                    for child in self.children:
                        child.draw_outer(zoom=zoom*child.scale, **kwargs)

                gl.glPushMatrix()
                gl.glColor3f(1, 1, 1)
                x, y, z = self.unlocked.obj_pos
                gl.glTranslatef(x, y, 0)
                gl.glScalef(1/64, 1/64, 0)
                for i, num in enumerate(state.key_config):
                    draw_string(str(int(num)), x=i*14/32*64+12, y=-5)
                gl.glPopMatrix()


    def unlock(self, key):
        from jrti_game.adjuster import Adjuster
        super().unlock(key)
        for i in range(4):
            a = Adjuster(
                parent=self,
                x=2+3/8+i*14/32,
                y=7+6/64,
                scale=1/128,
                width=24,
                height=80+24,
                instructions='Adjusting',
            )
            a.index = i


class Exclamation(Letter):
    keyhole = 1, 2.25, 2/32, 1/2
    innards = 1+1/8, 1+1/16, 2-3/16, 2-2/16

    def draw(self, zoom, **kwargs):
        super().draw(zoom=zoom, **kwargs)
        if self.unlocked:
            gl.glColor3f(0.1, 0.2, 0.5)
            with draw_opening(self):
                gl.glTranslatef(1, 1, 0)
                gl.glScalef(1/8, 1/8, 1)
                gl.glColor3f(0.4, 1, 0.2)
                sprites['eye_big'].blit(1.75, 1.5)


class Dot(Letter):
    keyhole = 1, 2.25, 2/32, 1/2
    innards = 1+1/8, 1+1/16, 2-3/16, 2-2/16

    def draw(self, zoom, **kwargs):
        super().draw(zoom=zoom, **kwargs)
        if self.unlocked:
            gl.glColor3f(0.1, 0.2, 0.5)
            with draw_opening(self):
                gl.glTranslatef(1, 1, 0)
                gl.glScalef(1/8, 1/8, 1)
                gl.glColor3f(0.4, 1, 0.2)
                sprites['bug_o'].blit(1.75, 1.5)


class Colon(Letter):
    keyhole = 1, 6.25, 2/32, 1/2
    innards = 1+1/8, 2+1/16, 2-3/16, 2-2/16

    def draw(self, zoom, **kwargs):
        super().draw(zoom=zoom, **kwargs)
        if self.unlocked:
            gl.glColor3f(0.1, 0.2, 0.5)
            with draw_opening(self):
                gl.glTranslatef(1, 2, 0)
                gl.glScalef(1/8, 1/8, 1)
                gl.glColor3f(0.4, 1, 0.2)
                sprites['hint_a'].blit(1.75, 1.5)

            gl.glTranslatef(0, 3, 0)
            gl.glColor3f(0.1, 0.2, 0.5)
            with draw_opening(self):
                gl.glTranslatef(1, 2, 0)
                gl.glScalef(1/8, 1/8, 1)
                gl.glColor3f(0.4, 1, 0.2)
                sprites['hint_b'].blit(1.75, 1.5)
            gl.glTranslatef(0, -3, 0)


SPECIAL_LETTER_CLASSES = {
    'i': Eye,
    'o': Oh,
    '!': Exclamation,
    '.': Dot,
    ':': Colon,
}



import os
import pyglet
from pyglet import gl
import ctypes
import math

import jrti_game.data
from jrti_game.data import sprites
import jrti_game.flippable
import jrti_game.text
from jrti_game.bug import BugArena
from jrti_game.end import Keyhole
from jrti_game.util import clamp
from jrti_game.state import state
from jrti_game import tool
from jrti_game.coords import mouse_to_logical
from jrti_game.coords import mouse_to_main_screen

window_style = getattr(
    pyglet.window.Window,
    'WINDOW_STYLE_' + os.environ.get('GAME_PYGLET_WINDOW_STYLE', 'DEFAULT'),
)

window = pyglet.window.Window(
    width=800,
    height=600,
    style=window_style,
)

main_screen = jrti_game.flippable.Layer(
    width=800,
    height=600,
    x=-400,
    y=-300,
    margin=32,
    instructions='End',
    instructions_color=(1, 0.9, 0.5),
)
state.main_screen = main_screen
bug_arena = BugArena(
    parent=main_screen,
    x=400 - 600//2,
    y=400 - 8*32,
    width=600,
    height=8*48,
    instructions='Main Plaque',
    margin=8,
    bgcolor=(.1, .1, .1, 1)
)

all(jrti_game.text.letters(
    "Just Read the", scale=4, y=8*36, x=300, center=True, parent=bug_arena))
all(jrti_game.text.letters(
    "Instructions!", scale=8, y=8*24, x=300, center=True, parent=bug_arena))

for x in range(2):
    bug_arena.spawn_bug()

all(jrti_game.text.letters(
    "They're behind everything!", scale=2, y=8*8, x=400, center=True,
    parent=main_screen))
all(jrti_game.text.letters(
    "PyWeek 24 entry", scale=1, y=8*3, x=400, center=True,
    parent=main_screen))
all(jrti_game.text.letters(
    "https://pyweek.org/e/instructions/", scale=1, y=8*1, x=400, center=True,
    parent=main_screen))

Keyhole(
    parent=main_screen,
    x=800-32-16*2, y=32,
    u=80-1, v=3*8, width=16+1, height=32+1, scale=2,
    instructions='Keyhole')

fps_display = pyglet.window.FPSDisplay(window)

pressed_keys = {}

state.tool = tool.Tool()


@window.event
def on_draw():
    window.clear()
    gl.glMatrixMode(gl.GL_PROJECTION)
    gl.glLoadMatrixf((ctypes.c_float * 16)(
        1, 0, 0, 0,
        0, 1, 0, 0,
        0, 0, 1, 1,
        0, 0, 0, 1))
    gl.glScalef(2/window.width, 2/window.height, 1/2048)
    gl.glMatrixMode(gl.GL_MODELVIEW)

    gl.glLoadIdentity()

    gl.glScalef(state.zoom, state.zoom, state.zoom)
    cx, cy = state.center
    gl.glTranslatef(-cx, -cy, 0)

    gl.glEnable(gl.GL_BLEND)
    gl.glBlendFunc(gl.GL_SRC_ALPHA, gl.GL_ONE_MINUS_SRC_ALPHA)
    gl.glEnable(gl.GL_TEXTURE_2D)
    gl.glBindTexture(gl.GL_TEXTURE_2D, jrti_game.data.spritesheet_texture.id)
    gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MAG_FILTER,
                       gl.GL_NEAREST)
    gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MIN_FILTER,
                       gl.GL_LINEAR)

    gl.glFrontFace(gl.GL_CCW)
    gl.glCullFace(gl.GL_BACK)
    gl.glDisable(gl.GL_CULL_FACE)

    main_screen.draw_outer(zoom=state.zoom)
    if state.tool.grabbing:
        state.tool.grabbing.draw_grab()
    main_screen.draw_flipping(zoom=state.zoom)

    gl.glLoadIdentity()
    gl.glTranslatef(state.tool_x, state.tool_y, 0)
    gl.glScalef(state.tool.scale, state.tool.scale, 1)
    state.tool.draw()

    if state.reassign_state:
        gl.glLoadIdentity()
        time0, sym1, time1, sym2, time2 = state.reassign_state
        t0 = clamp((state.time - (time0 or state.time)) * 5, 0, 1)
        t1 = clamp((state.time - (time1 or state.time)) * 5, 0, 1)
        t2 = clamp((state.time - (time2 or state.time)) * 5, 0, 1)
        t3 = clamp((state.time - (time2 or state.time)-4/10) * 5, 0, 1)
        t4 = clamp((state.time - (time2 or state.time)-7/10) * 5, 0, 1)
        gl.glLoadIdentity()
        gl.glTranslatef(0, 8, 0)
        gl.glColor3f(0.6, 0.5, 0.4)
        size = (t0-t4) * 8
        gl.glScalef(size, size, 0)
        gl.glPushMatrix()
        gl.glRotatef(t3*180, 0, 0, 1)
        sep = t1*8
        sprites['square'].blit(-sep, -10.5, width=sep*2, height=21)
        sprites['circ3'].blit(-11.5-sep, -11.5)
        sprites['circ3'].blit(-11.5+sep, -11.5)

        gl.glColor3f(0, 0, 0)
        if sym1:
            gl.glPushMatrix()
            gl.glTranslatef(-8, 0, 0)
            gl.glScalef(t1, t1, 0)
            gl.glRotatef(-t3*180, 0, 0, 1)
            jrti_game.text.draw_key(sym1)
            gl.glPopMatrix()
        if sym2:
            gl.glPushMatrix()
            gl.glTranslatef(8, 0, 0)
            gl.glScalef(t2, t2, 0)
            gl.glRotatef(-t3*180, 0, 0, 1)
            jrti_game.text.draw_key(sym2)
            gl.glPopMatrix()
        gl.glPopMatrix()
        gl.glScalef(1/8, 1/8, 0)
        if sym1:
            jrti_game.text.draw_string(state.keymap.get(sym1, ''),
                                       center=True, x=-8*8, y=-8*5)
        if sym2:
            jrti_game.text.draw_string(state.keymap.get(sym2, ''),
                                       center=True, x=8*8, y=-8*5)

    if os.environ.get('GAME_DEBUG'):
        gl.glLoadIdentity()
        fps_display.draw()


@window.event
def on_mouse_motion(x, y, button, mod):
    state.last_mouse_pos = x, y


@window.event
def on_mouse_press(x, y, button, mod):
    main_screen.mouse_release()
    if state.ending:
        return
    if mod & pyglet.window.key.MOD_SHIFT:
        state.alt_zoom_start = [y, (x, y)]
        return
    if mod & pyglet.window.key.MOD_CTRL:
        button = pyglet.window.mouse.RIGHT
    if button == pyglet.window.mouse.LEFT:
        if not state.reassign_state:
            main_screen.mouse_press(*mouse_to_main_screen(x, y),
                                    zoom=state.zoom)
            state.tool.deactivate()
    else:
        state.last_drag_pos = x, y


@window.event
def on_mouse_drag(x, y, dx, dy, button, mod):
    if state.ending:
        return
    if state.alt_zoom_start is not None:
        azs = state.alt_zoom_start
        zoom((y - azs[0]) / 2, *azs[1])
        state.alt_zoom_start[0] = y
    if state.last_drag_pos is not None:
        px, py = state.last_drag_pos
        cx, cy = state.center
        state.center = cx + (px - x) / state.zoom, cy + (py - y) / state.zoom
        finish_viewport_change()
        state.last_drag_pos = x, y
    else:
        main_screen.mouse_drag(*mouse_to_main_screen(x, y), zoom=state.zoom)
    state.last_mouse_pos = x, y


@window.event
def on_mouse_release(x, y, button, mod):
    state.alt_zoom_start = None
    state.last_drag_pos = None
    main_screen.mouse_release()


@window.event
def on_mouse_scroll(x, y, scroll_x, scroll_y):
    main_screen.mouse_release()
    if state.ending:
        return
    zoom(scroll_y, x, y)


@window.event
def on_key_press(sym, mod):
    # Apparently, some keyboards have an always-on modifier...
    # Only include the ones we care about.
    mod = mod & (pyglet.window.key.MOD_CTRL |
                 pyglet.window.key.MOD_SHIFT |
                 pyglet.window.key.MOD_ALT)

    if sym == pyglet.window.key.C and mod & pyglet.window.key.MOD_CTRL:
        raise KeyboardInterrupt()

    try:
        key = state.keymap[sym]
    except KeyError:
        key = ''

    if state.ending:
        if key in ('Exit', 'Quit'):
            exit()
        return pyglet.event.EVENT_HANDLED

    if state.reassign_state:
        main_screen.mouse_release()
        time0, sym1, time1, sym2, time2 = state.reassign_state
        if sym1 is None:
            state.reassign_state = time0, sym, state.time, None, None
        elif sym2 is None:
            state.reassign_state = time0, sym1, time1, sym, state.time
        return pyglet.event.EVENT_HANDLED

    if mod == 0:
        pressed_keys[key] = state.time

    if key in ('Exit', 'Quit'):
        exit()
    elif key == 'Grabby':
        if state.tool.name == 'grabby':
            _full_zoom()
            state.tool.grab()
        else:
            state.tool = tool.Grabby()
    elif key == 'Key':
        if state.tool.name == 'key':
            _full_zoom()
            state.tool.turn()
        else:
            state.tool = tool.Key()
    elif key == 'Reassign':
        if state.reassign_state is None:
            state.reassign_state = state.time, None, None, None, None
    elif key.startswith('Zoom '):
        try:
            level = int(key[5:])
        except ValueError:
            pass
        else:
            state.target_zoom = 1.3 ** level
    elif key == 'Cheat':
        print('Here, you cheater:', state.code)
        if mod & (pyglet.window.key.MOD_CTRL|pyglet.window.key.MOD_ALT):
            state.key_config[:] = state.code

    return pyglet.event.EVENT_HANDLED


@window.event
def on_key_release(sym, mod):
    try:
        key = state.keymap[sym]
    except KeyError:
        return
    pressed_keys.pop(key, None)


def zoom(amount, x=None, y=None):
    state.target_zoom = clamp(state.target_zoom * 1.1 ** amount,
                              state.min_zoom, 800)


def _zoom(amount):
    cx, cy = state.center
    lx, ly = mouse_to_logical(*state.last_mouse_pos)

    old = state.zoom
    state.zoom = amount

    cx = lx - (lx - cx) * old / state.zoom
    cy = ly - (ly - cy) * old / state.zoom
    state.center = cx, cy

    finish_viewport_change()


def _full_zoom():
    if state.zoom != state.target_zoom:
        _zoom(state.target_zoom)


def finish_viewport_change():
    cx, cy = state.center
    if state.zoom <= 1:
        cx, cy = 0, 0
    else:
        cx = clamp(cx, 400 - 400 * state.zoom, 400 * state.zoom - 400)
        cy = clamp(cy, 300 - 300 * state.zoom, 300 * state.zoom - 300)

    state.center = cx, cy

    state.tool.deactivate()


@pyglet.clock.schedule
def tick(dt):
    state.time += dt
    if state.zoom == state.target_zoom:
        pass
    elif 0.999 < state.zoom / state.target_zoom < 1.001:
        _zoom(state.target_zoom)
    else:
        x1 = state.zoom
        x2 = state.target_zoom
        w2 = 1 - math.exp(-dt) ** (10) ** state.zoom_easing
        w1 = 1 - w2
        _zoom((x1**w1 * x2**w2) ** (1/1))
    main_screen.tick(dt)

    tool_size = state.tool.size * state.tool.scale / 2
    tool_x = state.tool_x
    tool_y = state.tool_y
    for key, time in ((k, state.time - pressed_keys[k])
                      for k in pressed_keys):
        dz = .2 + (time*4)**3
        dm = clamp(2 + (time*6)**2, 0, 40)
        if key == 'Zoom Out':
            zoom(-dz * dt)
        elif key == 'Zoom In':
            zoom(dz * dt)
        elif key == 'Tool Left':
            tool_x = clamp(state.tool_x - dm, tool_size-400, 400-tool_size)
        elif key == 'Tool Right':
            tool_x = clamp(state.tool_x + dm, tool_size-400, 400-tool_size)
        elif key == 'Tool Down':
            tool_y = clamp(state.tool_y - dm, tool_size-300, 300-tool_size)
        elif key == 'Tool Up':
            tool_y = clamp(state.tool_y + dm, tool_size-300, 300-tool_size)
        elif key == 'View Left':
            cx, cy = state.center
            cx -= dm / state.zoom * dt * 15
            state.center = cx, cy
            finish_viewport_change()
        elif key == 'View Up':
            cx, cy = state.center
            cy += dm / state.zoom * dt * 15
            state.center = cx, cy
            finish_viewport_change()
        elif key == 'View Down':
            cx, cy = state.center
            cy -= dm / state.zoom * dt * 15
            state.center = cx, cy
            finish_viewport_change()
        elif key == 'View Right':
            cx, cy = state.center
            cx += dm / state.zoom * dt * 15
            state.center = cx, cy
            finish_viewport_change()

    if tool_x != state.tool_x or tool_y != state.tool_y:
        dx = tool_x - state.tool_x
        dy = tool_y - state.tool_y
        state.tool.move(dx, dy)
    else:
        state.tool.speed = 0

    if state.reassign_state:
        time0, sym1, time1, sym2, time2 = state.reassign_state
        if time2 and time2 < state.time - 9/10:
            state.keymap[sym1], state.keymap[sym2] = (
                state.keymap.get(sym2, ''), state.keymap.get(sym1, ''))
            main_screen.update_instructions()
            state.reassign_state = None

    state.tool.tick(dt)


def main():
    pyglet.app.run()



import textwrap
import pkgutil
from pathlib import Path
import tempfile
import os

import pyglet.image
import pyglet.font
import png


def get_data(name):
    return pkgutil.get_data('jrti_game', str(Path('data') / name))


with tempfile.TemporaryDirectory() as tmp:
    tmp_path = Path(tmp)

    def unpack_file(name):
        dest = tmp_path / name
        dest.parent.mkdir(parents=True, exist_ok=True)
        with open(str(dest), 'wb') as f:
            f.write(get_data(name))
        return str(dest)

    spritesheet_filename = unpack_file('spritesheet.png')
    spritesheet_image = pyglet.image.load(spritesheet_filename)
    spritesheet_texture = spritesheet_image.get_texture()

    def _load_spritesheet():
        with open(spritesheet_filename, 'rb') as f:
            width, height, data, metadata = png.Reader(f).read()
            data = list(data)
            data.reverse()
        return data, metadata['palette']

    spritesheet_data, spritesheet_palette = _load_spritesheet()

    pyglet.font.add_file(unpack_file('font/Itim-Regular.ttf'))
    instruction_font_name = "Itim"


def load_instructions():
    instruction_texts = {}

    current_text = []
    current_heading = ''

    def _load_instruction():
        text = textwrap.dedent('\n'.join(current_text)).strip()
        instruction_texts[current_heading] = text
        current_text.clear()

    for line in get_data('instructions.txt').decode().splitlines():
        if line.startswith('#'):
            _load_instruction()
            current_heading = line[1:].strip()
        else:
            current_text.append(line)
    _load_instruction()

    return instruction_texts


sprites = {
    'square': spritesheet_texture.get_region(0, 0, 1, 1),
    'eye': spritesheet_texture.get_region(0, 3*8, 9, 8),
    'crosshair': spritesheet_texture.get_region(10, 3*8, 8, 8),
    'ex': spritesheet_texture.get_region(18, 3*8, 8, 8),
    'bug': spritesheet_texture.get_region(28, 3*8+1, 11, 11),
    'bugleg': spritesheet_texture.get_region(40, 3*8+1, 7, 11),
    'grabby': spritesheet_texture.get_region(48, 3*8, 7, 7),
    'grabshadow': spritesheet_texture.get_region(48, 4*8+1, 8, 8),
    'grabtooth': spritesheet_texture.get_region(56, 3*8, 7, 7),
    'key': spritesheet_texture.get_region(63, 3*8, 18, 8),
    'key_head': spritesheet_texture.get_region(63, 3*8, 9, 8),
    'keyhole': spritesheet_texture.get_region(80, 3*8+1, 16, 32),
    'circ3': spritesheet_texture.get_region(96, 3*8+1, 23, 23),
    'eye_big': spritesheet_texture.get_region(0, 4*8+1, 13, 13),
    'bug_o': spritesheet_texture.get_region(0, 6*8+1, 13, 13),
    'hint_a': spritesheet_texture.get_region(16, 6*8+1, 13, 13),
    'hint_b': spritesheet_texture.get_region(32, 6*8+1, 13, 13),
    'adjuster': spritesheet_texture.get_region(48, 4*8+1, 23, 23),
}

default_keymap = {
    'LEFT': 'View Left',
    'UP': 'View Up',
    'DOWN': 'View Down',
    'RIGHT': 'View Right',
    'ESCAPE': 'Exit',
    'X': 'Quit',
    'R': 'Reassign',
    'G': 'Grabby',
    'K': 'Key',
    'O': 'Zoom Out',
    'P': 'Zoom In',
    'W': 'Tool Up',
    'A': 'Tool Left',
    'S': 'Tool Down',
    'D': 'Tool Right',

    '_0': 'Zoom 0',
    '_1': 'Zoom 1',
    '_2': 'Zoom 2',
    '_3': 'Zoom 3',
    '_4': 'Zoom 4',
    '_5': 'Zoom 5',
    '_6': 'Zoom 6',
    '_7': 'Zoom 7',
    '_8': 'Zoom 8',
    '_9': 'Zoom 9',
}

if os.environ.get('GAME_DEBUG'):
    default_keymap['C'] = 'Cheat'
